#include <SPI.h>
#include <Wire.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define OLED_RESET -1
#define SCREEN_ADDRESS 0x3C  //This is the I2C address for the screen.
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

//This is where the measured resistor values are used. Run the given calculation on the given resistors and plug the resulting value in.
//Example lines so you can get the syntax right: 
//int example = 123456789; 
//float example = 1.23456789;
//When calculating for the floating point variables, it is recommended to go to around 10 decimal places.
int N12VRi = ; //R35 Value in Ohms
int N12VRf = ; //R36 Value in Ohms
float MB12VCV = ;     //R47 / (R46 + R47)
float MB5VCV = ;      //R49 / (R48 + R49)
float MB33VCV = ;     //R53 / (R52 + R53)
float STBYCV = ;      //R55 / (R54 + R55)
float CPU12VCV = ;    //R45 / (R44 + R45)
float PCIE12VCV = ;   //R43 / (R42 + R43)
float PERIF12VCV = ;  //R41 / (R40 + R41)
float PERIF5VCV = ;   //R51 / (R50 + R51)
float PERIF33VCV = ;  //R57 / (R56 + R57)

//Measurement pins for PSU voltages
int MBN12VMP = 14;
int MB12VMP = 15;
int MB5VMP = 16;
int MB33VMP = 17;
int STBYMP = 20;
int CPU12VMP = 21;
int PCIE12VMP = 22;
int PERIF12VMP = 23;
int PERIF5VMP = 24;
int PERIF33VMP = 25;

//Other pins for control and sensing
int TOGGLELOAD = 26;
int PCISENSE = 27;
int psuPower = 28;
int psuLEDRed = 29;
int psuLEDGrn = 30;
int loadLED = 31;
int loadButton = 32;
int menuButton = 33;
int psuButton = 34;

//Global variables (some functions return to these)
int currentMenu = 0;
int PSUSpec;
boolean psuState = 0;
boolean loadState = 0;
boolean initialTestComplete = false;
boolean doNotRepeatEnableButton = true;
boolean loadTimerStarted = false;
boolean disableLoadButton = false;
unsigned long loadDisabledTime;
unsigned long loadStartedTime;
unsigned long voltageUpdatedTime;
unsigned long rippleUpdatedTime;

float MBN12V;
float MB12V;
float MB5V;
float MB33V;
float STBY;
float CPU12V;
float PCIE12V;
float PERIF12V;
float PERIF5V;
float PERIF33V;

float MBN12VRipRef;
float MB12VRipRef;
float MB5VRipRef;
float MB33VRipRef;
float STBYRipRef;
float CPU12VRipRef;
float PCIE12VRipRef;
float PERIF12VRipRef;
float PERIF5VRipRef;
float PERIF33VRipRef;

float MBN12VRipple;
float MB12VRipple;
float MB5VRipple;
float MB33VRipple;
float STBYRipple;
float CPU12VRipple;
float PCIE12VRipple;
float PERIF12VRipple;
float PERIF5VRipple;
float PERIF33VRipple;

float MBN12VSnapshots[3] = { 0,
                             0,
                             0 };

float MB12VSnapshots[3] = { 0,
                            0,
                            0 };

float MB5VSnapshots[3] = { 0,
                           0,
                           0 };

float MB33VSnapshots[3] = { 0,
                            0,
                            0 };

float STBYSnapshots[3] = { 0,
                           0,
                           0 };

float CPU12VSnapshots[3] = { 0,
                             0,
                             0 };

float PCIE12VSnapshots[3] = { 0,
                              0,
                              0 };

float PERIF12VSnapshots[3] = { 0,
                               0,
                               0 };

float PERIF5VSnapshots[3] = { 0,
                              0,
                              0 };

float PERIF33VSnapshots[3] = { 0,
                               0,
                               0 };

boolean MB12VPASS[3] = { false, false, false };
boolean MB5VPASS[3] = { false, false, false };
boolean MB33VPASS[3] = { false, false, false };
boolean MBN12VPASS[3] = { false, false, false };
boolean STBYPASS[3] = { false, false, false };
boolean CPU12VPASS[3] = { false, false, false };
boolean PCIE12VPASS[3] = { false, false, false };

void setup() {
  //Display initialization stuff
  Serial.begin(9600);
  if (!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for (;;)
      ;
  }
  display.clearDisplay();
  delay(250);
  display.display();
  delay(250);
  display.clearDisplay();

  analogReadResolution(12);

  pinMode(TOGGLELOAD, OUTPUT);
  pinMode(psuPower, OUTPUT);
  pinMode(psuLEDRed, OUTPUT);
  pinMode(psuLEDGrn, OUTPUT);
  pinMode(loadLED, OUTPUT);
  pinMode(PCISENSE, INPUT);
  pinMode(loadButton, INPUT);
  pinMode(menuButton, INPUT);
  pinMode(psuButton, INPUT);
  pinMode(MBN12VMP, INPUT);
  pinMode(MB12VMP, INPUT);
  pinMode(MB5VMP, INPUT);
  pinMode(MB33VMP, INPUT);
  pinMode(STBYMP, INPUT);
  pinMode(PCIE12VMP, INPUT);
  pinMode(CPU12VMP, INPUT);
  pinMode(PERIF12VMP, INPUT);
  pinMode(PERIF5VMP, INPUT);
  pinMode(PERIF33VMP, INPUT);

  //Startup menu and initial test start
  while (initialTestComplete == false) {
    initialScrn();
    togglePSUState(false);
    toggleLoadState(false);
    delay(100);
    if (digitalRead(psuButton) == HIGH) {
      initialTestComplete = false;
      break;
    }
    if (digitalRead(menuButton) == HIGH) {
      initialTestComplete = false;
      break;
    }
    if (digitalRead(loadButton) == HIGH) {
      initialTestComplete = false;
      break;
    }
  }
}




void loop() {
  //Enters the section of code that runs the initial test on the power supply
  if (initialTestComplete == false) {
restartTest:  //Goes here if an inconsistency error is found and user restarts the test by pressing one of the control buttons
    //This displays the testing screen and turns on the PSU/load
    testingScrn();
    togglePSUState(true);
    delay(1000);
    toggleLoadState(true);
    delay(250);
    //This takes our three snapshots of the PSU's voltages
    for (int i = 0; i < 3; i++) {
      testVoltages(1);
      MBN12VSnapshots[i] = MBN12V;
      MB12VSnapshots[i] = MB12V;
      MB5VSnapshots[i] = MB5V;
      MB33VSnapshots[i] = MB33V;
      STBYSnapshots[i] = STBY;
      CPU12VSnapshots[i] = CPU12V;
      PCIE12VSnapshots[i] = PCIE12V;
      PERIF12VSnapshots[i] = PERIF12V;
      PERIF5VSnapshots[i] = PERIF5V;
      PERIF33VSnapshots[i] = PERIF33V;
      delay(1500);
    }
    //This determines the correct spec that the plugged in PSU should be held to throughout testing, and also jumps to a failure screen if the spec is undeterminable.
    PSUSpec = determineAppropriateSpec();
    if (PSUSpec == 4) {
      undeterminableFailScrn();
      goto undeterminedFailure;
    }
    //This checks how the voltages captured in the snapshots compare to the specified voltages for the specific PSU we are testing. This makes it so that other functions can see what voltages/snapshots passed and what voltages/snapshots failed.
    checkSnapshots(PSUSpec);
    //This while loop puts the tester in a state where the PSU and load are off if there are inconsistencies found between the snapshots. This displays the correct screen and allows the user to either exit and restart the test, or exit and view the current voltage/ripple measurements on the unit anyway.
    while (searchInconsistencies(PSUSpec) == true) {
      togglePSUState(false);
      toggleLoadState(false);
      inconsistencyErrorScrn();
      if (digitalRead(menuButton) == HIGH) {
        goto seeVoltages;
      }
      if (digitalRead(psuButton) == HIGH) {
        goto restartTest;
      }
      if (digitalRead(loadButton) == HIGH) {
        goto restartTest;
      }
    }
    //Finds whether or not the PSU passed its test for its spec. If the function returns 0, it means it passed and the correct screen is displayed for the unit that has been detected.
    if (determinePassOrFail(PSUSpec) == 0) {
      if (PSUSpec == 1) {
        atxPassScrn();
      }
      if (PSUSpec == 2) {
        atxpciPassScrn();
      }
      if (PSUSpec == 3) {
        dellPassScrn();
      }
    }
    //If the pass/fail test does not return a zero, this uses the PSU spec to find the right screen then calls on the pass/fail test again to provide the failure number to the screen function. This will, in turn, display the correct failure screen for the PSU's spec and also display the exact failure that was found.
    if (determinePassOrFail(PSUSpec) != 0) {
      if (PSUSpec == 1) {
        atxFailScrn(determinePassOrFail(PSUSpec));
      }
      if (PSUSpec == 2) {
        atxpciFailScrn(determinePassOrFail(PSUSpec));
      }
      if (PSUSpec == 3) {
        dellFailScrn(determinePassOrFail(PSUSpec));
      }
    }
    //These markers are from goto's placed previously in the code. They allow those cases to bypass a lot of code and to reach the point where the user can navigate the menus and control the PSU/load.
undeterminedFailure:
seeVoltages:
    togglePSUState(false);
    toggleLoadState(false);
    initialTestComplete = true;
  }
  //Here can begin the code for how the tester behaves after it's done its initial test.
  //Set the current time each time the loop gets to this point again. This is used to keep track of the multiple things that need to be happening at different intervals.
  unsigned long currentTime = millis();

  displayCurrentMenu();  //Displays the current menu that is selected

  //Automatic shutdown of load after 25 seconds of it being on, and blocking of turing the load back on until 45 seconds of cooldown time has passed.
  if (loadState == true && loadTimerStarted == false) {
    loadStartedTime = millis();
    loadTimerStarted = true;
  }
  if (loadState == false) {
    loadTimerStarted = false;
  }
  if (currentTime >= loadStartedTime + 25000 && loadTimerStarted == true) {
    toggleLoadState(false);
    disableLoadButton = true;
    doNotRepeatEnableButton = false;
    loadDisabledTime = millis();
  }
  if (currentTime >= loadDisabledTime + 45000 && doNotRepeatEnableButton == false) {
    disableLoadButton = false;
    doNotRepeatEnableButton = true;
  }

  //Update voltage and ripple measurements periodically
  if (currentTime >= voltageUpdatedTime + 500) {
    testVoltages(1);
    voltageUpdatedTime = millis();
  }
  if (currentTime >= rippleUpdatedTime + 500) {
    findRipples(PSUSpec);
    rippleUpdatedTime = millis();
  }
  if (digitalRead(psuButton) == HIGH) {
    togglePSUState(!psuState);
    delay(250);
  }
  if (digitalRead(menuButton) == HIGH) {
    cycleMenu();
    delay(250);
  }
  if (digitalRead(loadButton) == HIGH && disableLoadButton == false) {
    toggleLoadState(!loadState);
    delay(250);
  }
}



//This is where the realm of the functions begins. Be careful, it gets scary down here (this all took forever to write)
void testVoltages(int type) {
  //Two types; type 1 will write the results to the global variables referenced by PSU testing and the voltage menus, and type 2 will post the results to variables referenced only by the ripple calculating code
  //Gather raw ADC values from the measure points
  int MBN12VADC;
  int MB12VADC;
  int MB5VADC;
  int MB33VADC;
  int STBYADC;
  int CPU12VADC;
  int PCIE12VADC;
  int PERIF12VADC;
  int PERIF5VADC;
  int PERIF33VADC;
  long MBN12VADCSUM = 0;
  long MB12VADCSUM = 0;
  long MB5VADCSUM = 0;
  long MB33VADCSUM = 0;
  long STBYADCSUM = 0;
  long CPU12VADCSUM = 0;
  long PCIE12VADCSUM = 0;
  long PERIF12VADCSUM = 0;
  long PERIF5VADCSUM = 0;
  long PERIF33VADCSUM = 0;
  int MBN12VADCAVG;
  int MB12VADCAVG;
  int MB5VADCAVG;
  int MB33VADCAVG;
  int STBYADCAVG;
  int CPU12VADCAVG;
  int PCIE12VADCAVG;
  int PERIF12VADCAVG;
  int PERIF5VADCAVG;
  int PERIF33VADCAVG;
  //Gather the raw ADC values by averaging 10 samlings to help eliminate noise/jitter that comes with the ESP32's non-ideal ADC.
  for (int i = 0; i < 10; i++) {
    MBN12VADC = analogRead(MBN12VMP);
    MB12VADC = analogRead(MB12VMP);
    MB5VADC = analogRead(MB5VMP);
    MB33VADC = analogRead(MB33VMP);
    STBYADC = analogRead(STBYMP);
    CPU12VADC = analogRead(CPU12VMP);
    PCIE12VADC = analogRead(PCIE12VMP);
    PERIF12VADC = analogRead(PERIF12VMP);
    PERIF5VADC = analogRead(PERIF5VMP);
    PERIF33VADC = analogRead(PERIF33VMP);
    MBN12VADCSUM = MBN12VADCSUM + MBN12VADC;
    MB12VADCSUM = MB12VADCSUM + MB12VADC;
    MB5VADCSUM = MB5VADCSUM + MB5VADC;
    MB33VADCSUM = MB33VADCSUM + MB33VADC;
    STBYADCSUM = STBYADCSUM + STBYADC;
    CPU12VADCSUM = CPU12VADCSUM + CPU12VADC;
    PCIE12VADCSUM = PCIE12VADCSUM + PCIE12VADC;
    PERIF12VADCSUM = PERIF12VADCSUM + PERIF12VADC;
    PERIF5VADCSUM = PERIF5VADCSUM + PERIF5VADC;
    PERIF33VADCSUM = PERIF33VADCSUM + PERIF33VADC;
  }
  MBN12VADCAVG = MBN12VADCSUM / 10;
  MB12VADCAVG = MB12VADCSUM / 10;
  MB5VADCAVG = MB5VADCSUM / 10;
  MB33VADCAVG = MB33VADCSUM / 10;
  STBYADCAVG = STBYADCSUM / 10;
  CPU12VADCAVG = CPU12VADCSUM / 10;
  PCIE12VADCAVG = PCIE12VADCSUM / 10;
  PERIF12VADCAVG = PERIF12VADCSUM / 10;
  PERIF5VADCAVG = PERIF5VADCSUM / 10;
  PERIF33VADCAVG = PERIF33VADCSUM / 10;
  float MBN12VRAW = MBN12VADCAVG * 0.0008056640625;
  float MB12VRAW = MB12VADCAVG * 0.0008056640625;
  float MB5VRAW = MB5VADCAVG * 0.0008056640625;
  float MB33VRAW = MB33VADCAVG * 0.0008056640625;
  float STBYRAW = STBYADCAVG * 0.0008056640625;
  float CPU12VRAW = CPU12VADCAVG * 0.0008056640625;
  float PCIE12VRAW = PCIE12VADCAVG * 0.0008056640625;
  float PERIF12VRAW = PERIF12VADCAVG * 0.0008056640625;
  float PERIF5VRAW = PERIF5VADCAVG * 0.0008056640625;
  float PERIF33VRAW = PERIF33VADCAVG * 0.0008056640625;
  //Convert measurement point voltage to the voltage at the input of the voltage dividers to get actual PSU voltage.
  //Reports final voltages to global variables, either the set referenced by the voltage screens or the set referenced by the ripple calculations
  if (type == 1) {
    float MBN12Vtemp = MBN12VRAW * N12VRi;
    MBN12V = MBN12Vtemp / N12VRf;
    MB12V = MB12VRAW / MB12VCV;
    MB5V = MB5VRAW / MB5VCV;
    MB33V = MB33VRAW / MB33VCV;
    STBY = STBYRAW / STBYCV;
    CPU12V = CPU12VRAW / CPU12VCV;
    PCIE12V = PCIE12VRAW / PCIE12VCV;
    PERIF12V = PERIF12VRAW / PERIF12VCV;
    PERIF5V = PERIF5VRAW / PERIF5VCV;
    PERIF33V = PERIF33VRAW / PERIF33VCV;
  }
  if (type == 2) {
    float MBN12Vtemp = MBN12VRAW * N12VRi;
    MBN12VRipRef = MBN12Vtemp / N12VRf;
    MB12VRipRef = MB12VRAW / MB12VCV;
    MB5VRipRef = MB5VRAW / MB5VCV;
    MB33VRipRef = MB33VRAW / MB33VCV;
    STBYRipRef = STBYRAW / STBYCV;
    CPU12VRipRef = CPU12VRAW / CPU12VCV;
    PCIE12VRipRef = PCIE12VRAW / PCIE12VCV;
    PERIF12VRipRef = PERIF12VRAW / PERIF12VCV;
    PERIF5VRipRef = PERIF5VRAW / PERIF5VCV;
    PERIF33VRipRef = PERIF33VRAW / PERIF33VCV;
  }
}

void findRipples(int spec) {
  //Finds the ripple in the given rail
  //Rail key: 1 - MB12V, 2 - MB5V, 3 - MB3.3V, 4 - STBY, 5 - MBN12V, 6 - CPU12V, 7 - PCIE12V, 8 - PERIF12V, 9 - PERIF5V, 10 - PERIF3.3V
  //Storage for the samples that the ripple is determined from. Room for 10 voltage samples on each rail. Also creates a variable that describes which rails are being measured.
  float MB12VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float MB5VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float MB33VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float STBYRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float MBN12VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float CPU12VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float PCIE12VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float PERIF12VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float PERIF5VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float PERIF33VRippleSamples[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  int rails[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  //Sets the data for what rails will be tested depending on PSU spec, as some PSUs do not have certain rails and therefore, should not have those rails tested
  if (spec == 1) {
    for (int i = 0; i < 6; i++) {
      rails[i] = 1;
    }
    for (int a = 7; a < 10; a++) {
      rails[a] = 1;
    }
  }
  if (spec == 2) {
    for (int i = 0; i < 10; i++) {
      rails[i] = 1;
    }
  }
  if (spec == 3) {
    rails[0] = 1;
    rails[3] = 1;
    rails[5] = 1;
  }
  if (spec == 4) {
    for (int i = 0; i < 10; i++) {
      rails[i] = 1;
    }
  }
  //Tests the rails and returns voltage values to the sample banks, where they are later processed into ripple measurements
  for (int i = 0; i < 10; i++) {
    testVoltages(2);
    if (rails[0] == 1) {
      MB12VRippleSamples[i] = MB12VRipRef;
    }
    if (rails[1] == 1) {
      MB5VRippleSamples[i] = MB5VRipRef;
    }
    if (rails[2] == 1) {
      MB33VRippleSamples[i] = MB33VRipRef;
    }
    if (rails[3] == 1) {
      STBYRippleSamples[i] = STBYRipRef;
    }
    if (rails[4] == 1) {
      MBN12VRippleSamples[i] = MBN12VRipRef;
    }
    if (rails[5] == 1) {
      CPU12VRippleSamples[i] = CPU12VRipRef;
    }
    if (rails[6] == 1) {
      PCIE12VRippleSamples[i] = PCIE12VRipRef;
    }
    if (rails[7] == 1) {
      PERIF12VRippleSamples[i] = PERIF12VRipRef;
    }
    if (rails[8] == 1) {
      PERIF5VRippleSamples[i] = PERIF5VRipRef;
    }
    if (rails[9] == 1) {
      PERIF33VRippleSamples[i] = PERIF33VRipRef;
    }
    delay(25);
  }
  //Here the banks of values are sent off to the findBiggestDifference(); function, which returns the biggest difference between two numbers in the inserted set. This value is the ripple value for the rail that corresponds to the sent values
  if (rails[0] == 1) {
    MB12VRipple = findBiggestDifference(MB12VRippleSamples);
  } else {
    MB12VRipple = 0;
  }
  if (rails[1] == 1) {
    MB5VRipple = findBiggestDifference(MB5VRippleSamples);
  } else {
    MB5VRipple = 0;
  }
  if (rails[2] == 1) {
    MB33VRipple = findBiggestDifference(MB33VRippleSamples);
  } else {
    MB33VRipple = 0;
  }
  if (rails[3] == 1) {
    STBYRipple = findBiggestDifference(STBYRippleSamples);
  } else {
    STBYRipple = 0;
  }
  if (rails[4] == 1) {
    MBN12VRipple = findBiggestDifference(MBN12VRippleSamples);
  } else {
    MBN12VRipple = 0;
  }
  if (rails[5] == 1) {
    CPU12VRipple = findBiggestDifference(CPU12VRippleSamples);
  } else {
    CPU12VRipple = 0;
  }
  if (rails[6] == 1) {
    PCIE12VRipple = findBiggestDifference(PCIE12VRippleSamples);
  } else {
    PCIE12VRipple = 0;
  }
  if (rails[7] == 1) {
    PERIF12VRipple = findBiggestDifference(PERIF12VRippleSamples);
  } else {
    PERIF12VRipple = 0;
  }
  if (rails[8] == 1) {
    PERIF5VRipple = findBiggestDifference(PERIF5VRippleSamples);
  } else {
    PERIF5VRipple = 0;
  }
  if (rails[9] == 1) {
    PERIF33VRipple = findBiggestDifference(PERIF33VRippleSamples);
  } else {
    PERIF33VRipple = 0;
  }
}

float findBiggestDifference(float samples[]) {
  float maxval = max10(samples);
  float minval = min10(samples);
  float difference = maxval - minval;
  return difference;
}

float max10(float numbers[10]) {
  //This function finds and returns the largest of 10 numbers
  float maxguess = 0;
  for (int i = 0; i < 10; i++) {
    maxguess = max(maxguess, numbers[i]);
  }
  return maxguess;
}

float min10(float numbers[10]) {
  //This function finds and returns the smallest of 10 numbers.
  float minguess = 9999;
  for (int i = 0; i < 10; i++) {
    minguess = min(minguess, numbers[i]);
  }
  return minguess;
}

int determineAppropriateSpec() {
  //Determines the appropriate spec to hold a power supply to based off of the stored voltage snapshot values.
  //Returns a number 1-4; 1 - ATX without PCIe Connection, 2 - ATX with PCIe, 3 - Dell proprietary, 4 - Undeterminable (happens with a non-PCI unit that is fully dead)
  if (scanPciPresence() == true) {  //If PCIe is connected, it has to be held to the ATX with PCIe spec, as that's the only spec that will have a PCIe cable connected.
    return 2;
  }
  if (MB5VSnapshots[1] < 0.05 && MB33VSnapshots[1] < 0.05 && PERIF12VSnapshots[1] < 0.05 && PERIF5VSnapshots[1] < 0.05 && PERIF33VSnapshots[1] < 0.05 && scanPciPresence() == false && STBYSnapshots[1] > 7.5) {  //If there is no 3.3V or 5V on the motherboard, no PCIe cable, and there is also no voltage coming from any peripheral connectors, it is likely a Dell unit.
    return 3;
  }
  if (STBYSnapshots[1] < 1 && MB5VSnapshots[1] < 0.05 && MB33VSnapshots[1] < 0.05 && PERIF12VSnapshots[1] < 0.05 && PERIF5VSnapshots[1] < 0.05 && PERIF33VSnapshots[1] < 0.05 && scanPciPresence() == false) {
    return 4;
  }
  if (scanPciPresence() == false) {  //If neither of the previous cases have been met, the function will reach this point, and since there's no PCIe cable, it's likely an ATX PSU with no PCIe cable.
    return 1;                        //If one of the other cases has already been met the function will return its value and quit, never reaching this point, so this shouldn't falsely classify any units even though its a condition both a Dell and non-PCIe ATX unit can technically meet.
  }
}

void checkSnapshots(int spec) {
  //This function checks the voltages gathered by the snapshots according to the found spec and posts the pass/fail results to global variables
  if (spec == 1) {
    for (int i = 0; i < 3; i++) {
      if (MB12VSnapshots[i] > 11.4 && MB12VSnapshots[i] < 12.6) {
        MB12VPASS[i] = true;
      } else {
        MB12VPASS[i] = false;
      }
      if (MB5VSnapshots[i] > 4.75 && MB5VSnapshots[i] < 5.25) {
        MB5VPASS[i] = true;
      } else {
        MB5VPASS[i] = false;
      }
      if (MB33VSnapshots[i] > 3.135 && MB33VSnapshots[i] < 3.465) {
        MB33VPASS[i] = true;
      } else {
        MB33VPASS[i] = false;
      }
      if (MBN12VSnapshots[i] > 10.8 && MBN12VSnapshots[i] < 13.2) {
        MBN12VPASS[i] = true;
      } else {
        MBN12VPASS[i] = false;
      }
      if (STBYSnapshots[i] > 4.75 && STBYSnapshots[i] < 5.25) {
        STBYPASS[i] = true;
      } else {
        STBYPASS[i] = false;
      }
      if (CPU12VSnapshots[i] > 11.4 && CPU12VSnapshots[i] < 12.6) {
        CPU12VPASS[i] = true;
      } else {
        CPU12VPASS[i] = false;
      }
    }
  }
  if (spec == 2) {
    for (int i = 0; i < 3; i++) {
      if (MB12VSnapshots[i] > 11.4 && MB12VSnapshots[i] < 12.6) {
        MB12VPASS[i] = true;
      } else {
        MB12VPASS[i] = false;
      }
      if (MB5VSnapshots[i] > 4.75 && MB5VSnapshots[i] < 5.25) {
        MB5VPASS[i] = true;
      } else {
        MB5VPASS[i] = false;
      }
      if (MB33VSnapshots[i] > 3.135 && MB33VSnapshots[i] < 3.465) {
        MB33VPASS[i] = true;
      } else {
        MB33VPASS[i] = false;
      }
      if (MBN12VSnapshots[i] > 10.8 && MBN12VSnapshots[i] < 13.2) {
        MBN12VPASS[i] = true;
      } else {
        MBN12VPASS[i] = false;
      }
      if (STBYSnapshots[i] > 4.75 && STBYSnapshots[i] < 5.25) {
        STBYPASS[i] = true;
      } else {
        STBYPASS[i] = false;
      }
      if (CPU12VSnapshots[i] > 11.4 && CPU12VSnapshots[i] < 12.6) {
        CPU12VPASS[i] = true;
      } else {
        CPU12VPASS[i] = false;
      }
      if (PCIE12VSnapshots[i] > 11.4 && PCIE12VSnapshots[i] < 12.6) {
        PCIE12VPASS[i] = true;
      } else {
        PCIE12VPASS[i] = false;
      }
    }
  }
  if (spec == 3) {
    for (int i = 0; i < 3; i++) {
      if (MB12VSnapshots[i] > 11.4 && MB12VSnapshots[i] < 12.6) {
        MB12VPASS[i] = true;
      } else {
        MB12VPASS[i] = false;
      }
      if (STBYSnapshots[i] > 11.4 && STBYSnapshots[i] < 12.6) {
        STBYPASS[i] = true;
      } else {
        STBYPASS[i] = false;
      }
      if (CPU12VSnapshots[i] > 11.4 && CPU12VSnapshots[i] < 12.6) {
        CPU12VPASS[i] = true;
      } else {
        CPU12VPASS[i] = false;
      }
    }
  }
}

boolean searchInconsistencies(int spec) {
  //Returns either True or False; True = Inconsistencies in the test samples were present, False = No inconsistencies were found.
  if (spec == 1) {
    boolean MB12VConsistent = false;
    boolean MB5VConsistent = false;
    boolean MB33VConsistent = false;
    boolean MBN12VConsistent = false;
    boolean STBYConsistent = false;
    boolean CPU12VConsistent = false;
    if (MB12VPASS[0] == MB12VPASS[1] && MB12VPASS[0] == MB12VPASS[2] && MB12VPASS[1] == MB12VPASS[2]) {
      MB12VConsistent = true;
    } else {
      MB12VConsistent = false;
    }
    if (MB5VPASS[0] == MB5VPASS[1] && MB5VPASS[0] == MB5VPASS[2] && MB5VPASS[1] == MB5VPASS[2]) {
      MB5VConsistent = true;
    } else {
      MB5VConsistent = false;
    }
    if (MB33VPASS[0] == MB33VPASS[1] && MB33VPASS[0] == MB33VPASS[2] && MB33VPASS[1] == MB33VPASS[2]) {
      MB33VConsistent = true;
    } else {
      MB33VConsistent = false;
    }
    if (MBN12VPASS[0] == MBN12VPASS[1] && MBN12VPASS[0] == MBN12VPASS[2] && MBN12VPASS[1] == MBN12VPASS[2]) {
      MBN12VConsistent = true;
    } else {
      MBN12VConsistent = false;
    }
    if (STBYPASS[0] == STBYPASS[1] && STBYPASS[0] == STBYPASS[2] && STBYPASS[1] == STBYPASS[2]) {
      STBYConsistent = true;
    } else {
      STBYConsistent = false;
    }
    if (CPU12VPASS[0] == CPU12VPASS[1] && CPU12VPASS[0] == CPU12VPASS[2] && CPU12VPASS[1] == CPU12VPASS[2]) {
      CPU12VConsistent = true;
    } else {
      CPU12VConsistent = false;
    }
    if (MB12VConsistent == true && MB5VConsistent == true && MB33VConsistent == true && MBN12VConsistent == true && STBYConsistent == true && CPU12VConsistent == true) {
      return false;
    } else {
      return true;
    }
  }
  if (spec == 2) {
    boolean MB12VConsistent = false;
    boolean MB5VConsistent = false;
    boolean MB33VConsistent = false;
    boolean MBN12VConsistent = false;
    boolean STBYConsistent = false;
    boolean CPU12VConsistent = false;
    boolean PCIE12VConsistent = false;
    if (MB12VPASS[0] == MB12VPASS[1] && MB12VPASS[0] == MB12VPASS[2] && MB12VPASS[1] == MB12VPASS[2]) {
      MB12VConsistent = true;
    } else {
      MB12VConsistent = false;
    }
    if (MB5VPASS[0] == MB5VPASS[1] && MB5VPASS[0] == MB5VPASS[2] && MB5VPASS[1] == MB5VPASS[2]) {
      MB5VConsistent = true;
    } else {
      MB5VConsistent = false;
    }
    if (MB33VPASS[0] == MB33VPASS[1] && MB33VPASS[0] == MB33VPASS[2] && MB33VPASS[1] == MB33VPASS[2]) {
      MB33VConsistent = true;
    } else {
      MB33VConsistent = false;
    }
    if (MBN12VPASS[0] == MBN12VPASS[1] && MBN12VPASS[0] == MBN12VPASS[2] && MBN12VPASS[1] == MBN12VPASS[2]) {
      MBN12VConsistent = true;
    } else {
      MBN12VConsistent = false;
    }
    if (STBYPASS[0] == STBYPASS[1] && STBYPASS[0] == STBYPASS[2] && STBYPASS[1] == STBYPASS[2]) {
      STBYConsistent = true;
    } else {
      STBYConsistent = false;
    }
    if (CPU12VPASS[0] == CPU12VPASS[1] && CPU12VPASS[0] == CPU12VPASS[2] && CPU12VPASS[1] == CPU12VPASS[2]) {
      CPU12VConsistent = true;
    } else {
      CPU12VConsistent = false;
    }
    if (PCIE12VPASS[0] == PCIE12VPASS[1] && PCIE12VPASS[0] == PCIE12VPASS[2] && PCIE12VPASS[1] == PCIE12VPASS[2]) {
      PCIE12VConsistent = true;
    } else {
      PCIE12VConsistent = false;
    }
    if (MB12VConsistent == true && MB5VConsistent == true && MB33VConsistent == true && MBN12VConsistent == true && STBYConsistent == true && CPU12VConsistent == true && PCIE12VConsistent == true) {
      return false;
    } else {
      return true;
    }
  }
  if (spec == 3) {
    boolean MB12VConsistent = false;
    boolean STBYConsistent = false;
    boolean CPU12VConsistent = false;
    if (MB12VPASS[0] == MB12VPASS[1] && MB12VPASS[0] == MB12VPASS[2] && MB12VPASS[1] == MB12VPASS[2]) {
      MB12VConsistent = true;
    } else {
      MB12VConsistent = false;
    }
    if (STBYPASS[0] == STBYPASS[1] && STBYPASS[0] == STBYPASS[2] && STBYPASS[1] == STBYPASS[2]) {
      STBYConsistent = true;
    } else {
      STBYConsistent = false;
    }
    if (CPU12VPASS[0] == CPU12VPASS[1] && CPU12VPASS[0] == CPU12VPASS[2] && CPU12VPASS[1] == CPU12VPASS[2]) {
      CPU12VConsistent = true;
    } else {
      CPU12VConsistent = false;
    }
    if (MB12VConsistent == true && STBYConsistent == true && CPU12VConsistent == true) {
      return false;
    } else {
      return true;
    }
  }
}

int determinePassOrFail(int spec) {
  //Returns an integer from 0-7 indicating whether the PSU passed or not, and if not, the reason that the power supply unit failed. This reason can be plugged into the failure notification screen function to display the correct screen.
  //0 == Unit Passed
  if (spec == 1) {
    //If unit passes, quit and return 0 indicating a pass
    if (MB12VPASS[1] == true && MB5VPASS[1] == true && MB33VPASS[1] == true && MBN12VPASS[1] == true && STBYPASS[1] == true && CPU12VPASS[1] == true) {
      return 0;
    }
    //Establish an array where the discovered failures are saved temporarily
    int failureTracker[6] = { 0, 0, 0, 0, 0, 0 };
    //Check for failure #1 - Complete voltage failure (excluding STBY)
    if (MB12VPASS[1] == false && MB5VPASS[1] == false && MB33VPASS[1] == false && MBN12VPASS[1] == false && CPU12VPASS[1] == false) {
      failureTracker[0] = 1;
    }
    //Check for failure #2 - Motherboard voltage failure
    if (MB12VPASS[1] == false or MB5VPASS[1] == false or MB33VPASS[1] == false or MBN12VPASS[1] == false) {
      failureTracker[1] = 1;
    }
    //Check for failure #3 - STBY voltage failure
    if (STBYPASS[1] == false) {
      failureTracker[2] = 1;
    }
    //Check for failure #4 - CPU voltage failure
    if (CPU12VPASS[1] == false) {
      failureTracker[3] = 1;
    }
    //Skip check for failure #5 - PCIe voltage failure, as this spec does not have any PCIe cable attached and should not be failed for a lack of PCIe voltage.
    failureTracker[4] = 0;
    //Calculate ripple on the rails for this spec and ensure that the measured ripple is not too high.
    findRipples(1);
    if (MB12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (MB5VRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (MB33VRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (MBN12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (STBYRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (CPU12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    //Check to see if a full voltage failure was found, in which case this function can return that information and quit in order to not flag as "multiple voltage failures"
    if (failureTracker[0] == 1) {
      return 1;
    }
    //If multiple failures are detected, excluding a full voltage failure which will already have been reported before the code reaches this point, return the multiple voltage failure value
    if (failureTracker[1] + failureTracker[2] + failureTracker[3] + failureTracker[4] + failureTracker[5] > 1) {
      return 6;
    }
    //Return corresponding values for voltage failures 2, 3, 4, 5, and 7 if present.
    if (failureTracker[1] == 1) {
      return 2;
    }
    if (failureTracker[2] == 1) {
      return 3;
    }
    if (failureTracker[3] == 1) {
      return 4;
    }
    if (failureTracker[4] == 1) {
      return 5;
    }
    if (failureTracker[5] == 1) {
      return 7;
    }
  }
  if (spec == 2) {
    if (MB12VPASS[1] == true && MB5VPASS[1] == true && MB33VPASS[1] == true && MBN12VPASS[1] == true && STBYPASS[1] == true && CPU12VPASS[1] == true && PCIE12VPASS[1] == true) {
      return 0;
    }
    //Establish an array where the discovered failures are saved temporarily
    int failureTracker[6] = { 0, 0, 0, 0, 0, 0 };
    //Check for failure #1 - Complete voltage failure (excluding STBY)
    if (MB12VPASS[1] == false && MB5VPASS[1] == false && MB33VPASS[1] == false && MBN12VPASS[1] == false && CPU12VPASS[1] == false && PCIE12VPASS[1] == false) {
      failureTracker[0] = 1;
    }
    //Check for failure #2 - Motherboard voltage failure
    if (MB12VPASS[1] == false or MB5VPASS[1] == false or MB33VPASS[1] == false or MBN12VPASS[1] == false) {
      failureTracker[1] = 1;
    }
    //Check for failure #3 - STBY voltage failure
    if (STBYPASS[1] == false) {
      failureTracker[2] = 1;
    }
    //Check for failure #4 - CPU voltage failure
    if (CPU12VPASS[1] == false) {
      failureTracker[3] = 1;
    }
    //Check for failure #5 - PCIe voltage failure
    if (PCIE12VPASS[1] == false) {
      failureTracker[4] = 1;
    }
    //Calculate ripple on the rails for this spec and ensure that the measured ripple is not too high.
    findRipples(2);
    if (MB12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (MB5VRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (MB33VRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (MBN12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (STBYRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (CPU12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (PCIE12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    //Check to see if a full voltage failure was found, in which case this function can return that information and quit in order to not flag as "multiple voltage failures"
    if (failureTracker[0] == 1) {
      return 1;
    }
    //If multiple failures are detected, excluding a full voltage failure which will already have been reported before the code reaches this point, return the multiple voltage failure value
    if (failureTracker[1] + failureTracker[2] + failureTracker[3] + failureTracker[4] + failureTracker[5] > 1) {
      return 6;
    }
    //Return corresponding values for voltage failures 2, 3, 4, and 5 if present.
    if (failureTracker[1] == 1) {
      return 2;
    }
    if (failureTracker[2] == 1) {
      return 3;
    }
    if (failureTracker[3] == 1) {
      return 4;
    }
    if (failureTracker[4] == 1) {
      return 5;
    }
    if (failureTracker[5] == 1) {
      return 7;
    }
  }
  if (spec == 3) {
    if (MB12VPASS[1] == true && STBYPASS[1] == true && CPU12VPASS[1] == true) {
      return 0;
    }
    //Establish an array where the discovered failures are saved temporarily
    int failureTracker[6] = { 0, 0, 0, 0, 0, 0 };
    //Check for failure #1 - Complete voltage failure (excluding STBY)
    if (MB12VPASS[1] == false && CPU12VPASS[1] == false) {
      failureTracker[0] = 1;
    }
    //Check for failure #2 - Motherboard voltage failure
    if (MB12VPASS[1] == false) {
      failureTracker[1] = 1;
    }
    //Check for failure #3 - STBY voltage failure
    if (STBYPASS[1] == false) {
      failureTracker[2] = 1;
    }
    //Check for failure #4 - CPU voltage failure
    if (CPU12VPASS[1] == false) {
      failureTracker[3] = 1;
    }
    //Skip check for failure #5 - PCIe voltage failure, as this spec does not have any PCIe cable attached and should not be failed for a lack of PCIe voltage.
    failureTracker[4] = 0;
    //Calculate ripple on the rails for this spec and ensure that the measured ripple is not too high.
    findRipples(3);
    if (MB12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    if (STBYRipple > 0.05) {
      failureTracker[5] = 1;
    }
    if (CPU12VRipple > 0.12) {
      failureTracker[5] = 1;
    }
    //Check to see if a full voltage failure was found, in which case this function can return that information and quit in order to not flag as "multiple voltage failures"
    if (failureTracker[0] == 1) {
      return 1;
    }
    //If multiple failures are detected, excluding a full voltage failure which will already have been reported before the code reaches this point, return the multiple voltage failure value
    if (failureTracker[1] + failureTracker[2] + failureTracker[3] + failureTracker[4] + failureTracker[5] > 1) {
      return 6;
    }
    //Return corresponding values for voltage failures 2, 3, 4, and 5 if present.
    if (failureTracker[1] == 1) {
      return 2;
    }
    if (failureTracker[2] == 1) {
      return 3;
    }
    if (failureTracker[3] == 1) {
      return 4;
    }
    if (failureTracker[4] == 1) {
      return 5;
    }
    if (failureTracker[5] == 1) {
      return 7;
    }
  }
}

boolean scanPciPresence() {
  //Returns either True or False; True = PCIe connector is connected, False = No PCIe connector is connected
  if (digitalRead(PCISENSE) == HIGH) {
    return false;
  } else {
    return true;
  }
}

void cycleMenu() {
  //Changes the displayed menu on the OLED display
  if (currentMenu == 4) {
    currentMenu = 1;
    return;
  }
  if (currentMenu < 4) {
    currentMenu++;
  }
  displayCurrentMenu();
}

void displayCurrentMenu() {
  if (currentMenu == 1) {
    voltagesPageOne();
  }
  if (currentMenu == 2) {
    voltagesPageTwo();
  }
  if (currentMenu == 3) {
    ripplePageOne(PSUSpec);
  }
  if (currentMenu == 4) {
    ripplePageTwo(PSUSpec);
  }
}

void toggleLoadState(boolean state) {
  //boolean "status" is either True or False; True = Load Enabled, False = Load Disabled
  //Returns the current status of the load to the global variable "loadState"
  if (state == true) {
    digitalWrite(TOGGLELOAD, HIGH);
    digitalWrite(loadLED, HIGH);
    loadState = 1;
  }
  if (state == false) {
    digitalWrite(TOGGLELOAD, LOW);
    digitalWrite(loadLED, LOW);
    loadState = 0;
  }
}

void togglePSUState(boolean state) {
  //boolean "status" is either True or False; True = PSU ON, False = PSU OFF
  //Returns the current status of the PSU to the globar variable "psuState"
  if (state == true) {
    digitalWrite(psuPower, HIGH);
    digitalWrite(psuLEDRed, HIGH);
    digitalWrite(psuLEDGrn, LOW);
    psuState = 1;
  }
  if (state == false) {
    digitalWrite(psuPower, LOW);
    digitalWrite(psuLEDRed, LOW);
    digitalWrite(psuLEDGrn, HIGH);
    psuState = 0;
  }
}

void initialScrn() {
  display.setTextSize(2);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("Turn on");
  display.println("the PSU to");
  display.println("begin the");
  display.println("test.");
  display.display();
}

void testingScrn() {
  display.clearDisplay();
  display.setTextSize(2);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("Testing...");
  display.display();
}

void inconsistencyErrorScrn() {
  display.clearDisplay();
  display.setTextSize(2);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("FAILURE:");
  display.setTextSize(1);
  display.println("Inconsistent");
  display.println("test results");
  display.println("across samplings");
  display.println("PSU has FAILED.");
  display.display();
}

void atxPassScrn() {
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("PASS");
  display.setTextSize(1);
  display.println("Spec: ATX");
  display.println("No issues found.");
  display.display();
}

void atxpciPassScrn() {
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("PASS");
  display.setTextSize(1);
  display.println("Spec: ATX (PCIe)");
  display.println("No issues found.");
  display.display();
}

void atxFailScrn(int reason) {
  //Reason key: 1 - Full Failure (ALL bad voltages, ignores STBY), 2 - Mobo Voltage Failure, 3 - STBY Voltage Failure, 4 - CPU Voltage Failure, 5 - PCIe Voltage Failure, 6 - Multiple Voltage Failures, 7 - Extreme Ripple
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("FAIL");
  display.setTextSize(1);
  display.println("Spec: ATX");
  if (reason == 1) {
    display.println("Complete voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 2) {
    display.println("Mobo voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 3) {
    display.println("STBY voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 4) {
    display.println("CPU voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 5) {
    display.println("PCIe voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 6) {
    display.println("Multiple voltage");
    display.println("or ripple failures.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 7) {
    display.println("Extreme ripple.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
}

void atxpciFailScrn(int reason) {
  //Reason key: 1 - Full Failure (ALL bad voltages, ignores STBY), 2 - Mobo Voltage Failure, 3 - STBY Voltage Failure, 4 - CPU Voltage Failure, 5 - PCIe Voltage Failure, 6 - Multiple Voltage Failures, 7 - Extreme Ripple
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("FAIL");
  display.setTextSize(1);
  display.println("Spec: ATX (PCIe)");
  if (reason == 1) {
    display.println("Complete voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 2) {
    display.println("Mobo voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 3) {
    display.println("STBY voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 4) {
    display.println("CPU voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 5) {
    display.println("PCIe voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 6) {
    display.println("Multiple voltage");
    display.println("or ripple failures.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 7) {
    display.println("Extreme ripple.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
}

void dellPassScrn() {
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("PASS");
  display.setTextSize(1);
  display.println("Spec: Dell");
  display.println("Proprietary");
  display.println("No issues found.");
  display.display();
}



void dellFailScrn(int reason) {
  //Reason key: 1 - Full Failure (ALL bad voltages, ignores STBY), 2 - Mobo Voltage Failure, 3 - STBY Voltage Failure, 4 - CPU Voltage Failure, 5 - PCIe Voltage Failure (N/A on Dell Units), 6 - Multiple Voltage Failures, 7 - Extreme Ripple
  display.clearDisplay();
  display.setTextSize(3);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("FAIL");
  display.setTextSize(1);
  display.println("Spec: Dell");
  if (reason == 1) {
    display.println("Complete voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 2) {
    display.println("Mobo voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 3) {
    display.println("STBY voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 4) {
    display.println("CPU voltage");
    display.println("failure.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 5) {
    display.println("Error. Restart");
    display.println("tester now.");
    return;
  }
  if (reason == 6) {
    display.println("Multiple voltage");
    display.println("or ripple failures.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
  if (reason == 7) {
    display.println("Extreme ripple.");
    display.println("\"Menu\" for voltages");
    display.display();
    return;
  }
}

void undeterminableFailScrn() {
  display.clearDisplay();
  display.setTextSize(2);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println("FAIL");
  display.setTextSize(1);
  display.println("Spec: Undetermined");
  display.println("Unit could not be");
  display.println("identified. This");
  display.println("means it failed");
  display.println("catastrophically.");
  display.println("\"Menu\" for voltages");
  display.display();
}

void voltagesPageOne() {
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.print("MB 12V: ");
  display.println(MB12V);
  display.print("MB 5V: ");
  display.println(MB5V);
  display.print("MB 3.3V: ");
  display.println(MB33V);
  display.print("CPU 12V: ");
  display.println(CPU12V);
  display.print("PCIe 12V: ");
  display.println(PCIE12V);
  display.display();
}

void voltagesPageTwo() {
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.print("-12V: -");
  display.println(MBN12V);
  display.print("STBY: ");
  display.println(STBY);
  display.print("Perif 12V: ");
  display.println(PERIF12V);
  display.print("Perif 5V: ");
  display.println(PERIF5V);
  display.print("Perif 3.3V: ");
  display.println(PERIF33V);
  display.display();
}

void ripplePageOne(int spec) {
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  float MB12VRippleMV = MB12VRipple * 1000;
  float MB5VRippleMV = MB5VRipple * 1000;
  float MB33VRippleMV = MB33VRipple * 1000;
  float CPU12VRippleMV = CPU12VRipple * 1000;
  float PCIE12VRippleMV = PCIE12VRipple * 1000;
  if (spec == 1) {
    display.println("Ripple Measurements");
    display.print("MB 12V: ");
    display.print(MB12VRippleMV);
    display.println("mV");
    display.print("MB 5V: ");
    display.print(MB5VRippleMV);
    display.println("mV");
    display.print("MB 3.3V: ");
    display.print(MB33VRippleMV);
    display.println("mV");
    display.print("CPU 12V: ");
    display.print(CPU12VRippleMV);
    display.println("mV");
    display.print("PCIE 12V: ");
    display.print("N/A");
    display.display();
    return;
  }
  if (spec == 2) {
    display.println("Ripple Measurements");
    display.print("MB 12V: ");
    display.print(MB12VRippleMV);
    display.println("mV");
    display.print("MB 5V: ");
    display.print(MB5VRippleMV);
    display.println("mV");
    display.print("MB 3.3V: ");
    display.print(MB33VRippleMV);
    display.println("mV");
    display.print("CPU 12V: ");
    display.print(CPU12VRippleMV);
    display.println("mV");
    display.print("PCIE 12V: ");
    display.print(PCIE12VRippleMV);
    display.print("mv");
    display.display();
    return;
  }
  if (spec == 3) {
    display.println("Ripple Measurements");
    display.print("MB 12V: ");
    display.print(MB12VRippleMV);
    display.println("mV");
    display.print("MB 5V: ");
    display.println("N/A");
    display.print("MB 3.3V: ");
    display.println("N/A");
    display.print("CPU 12V: ");
    display.print(CPU12VRippleMV);
    display.println("mV");
    display.print("PCIE 12V: ");
    display.print("N/A");
    display.display();
    return;
  }
  if (spec == 4) {
    display.println("Ripple Measurements");
    display.print("MB 12V: ");
    display.print(MB12VRippleMV);
    display.println("mV");
    display.print("MB 5V: ");
    display.print(MB5VRippleMV);
    display.println("mV");
    display.print("MB 3.3V: ");
    display.print(MB33VRippleMV);
    display.println("mV");
    display.print("CPU 12V: ");
    display.print(CPU12VRippleMV);
    display.println("mV");
    display.print("PCIE 12V: ");
    display.print(PCIE12VRippleMV);
    display.print("mv");
    display.display();
    return;
  }
}

void ripplePageTwo(int spec) {
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  float STBYRippleMV = STBYRipple * 1000;
  float MBN12VRippleMV = MBN12VRipple * 1000;
  float PERIF12VRippleMV = PERIF12VRipple * 1000;
  float PERIF5VRippleMV = PERIF5VRipple * 1000;
  float PERIF33VRippleMV = PERIF33VRipple * 1000;
  if (spec == 1) {
    display.println("Ripple Measurements");
    display.print("STBY: ");
    display.print(STBYRippleMV);
    display.println("mV");
    display.print("MB -12V: ");
    display.print(MBN12VRippleMV);
    display.println("mV");
    display.print("Perif. 12V: ");
    display.print(PERIF12VRippleMV);
    display.println("mV");
    display.print("Perif. 5V: ");
    display.print(PERIF5VRippleMV);
    display.println("mV");
    display.print("Perif. 3.3V: ");
    display.print(PERIF33VRippleMV);
    display.print("mv");
    display.display();
  }
  if (spec == 2) {
    display.println("Ripple Measurements");
    display.print("STBY: ");
    display.print(STBYRippleMV);
    display.println("mV");
    display.print("MB -12V: ");
    display.print(MBN12VRippleMV);
    display.println("mV");
    display.print("Perif. 12V: ");
    display.print(PERIF12VRippleMV);
    display.println("mV");
    display.print("Perif. 5V: ");
    display.print(PERIF5VRippleMV);
    display.println("mV");
    display.print("Perif. 3.3V: ");
    display.print(PERIF33VRippleMV);
    display.print("mv");
    display.display();
  }
  if (spec == 3) {
    display.println("Ripple Measurements");
    display.print("STBY: ");
    display.print(STBYRippleMV);
    display.println("mV");
    display.print("MB -12V: ");
    display.println("N/A");
    display.print("Perif. 12V: ");
    display.println("N/A");
    display.print("Perif. 5V: ");
    display.println("N/A");
    display.print("Perif. 3.3V: ");
    display.print("N/A");
    display.display();
  }
  if (spec == 4) {
    display.println("Ripple Measurements");
    display.print("STBY: ");
    display.print(STBYRippleMV);
    display.println("mV");
    display.print("MB -12V: ");
    display.print(MBN12VRippleMV);
    display.println("mV");
    display.print("Perif. 12V: ");
    display.print(PERIF12VRippleMV);
    display.println("mV");
    display.print("Perif. 5V: ");
    display.print(PERIF5VRippleMV);
    display.println("mV");
    display.print("Perif. 3.3V: ");
    display.print(PERIF33VRippleMV);
    display.print("mv");
    display.display();
  }
}